import json
from jsonschema import validate

schema = {
    "type": "object",
    "properties": {
        "identifier": {"type": "string"},
        "type": {"type": "string", "enum": ["list", "instance", "chart"]},
        "endpoints": {
            "type": "object",
            "properties": {
                "list": {
                    "type": "string",
                    "format": "uri",
                    "pattern": "^(https?|http?|wss?|ws?|ftp)://",
                },
                "instance": {
                    "type": "string",
                    "format": "uri",
                    "pattern": "^(https?|http?|wss?|ws?|ftp)://",
                },
                "new": {
                    "type": "string",
                    "format": "uri",
                    "pattern": "^(https?|http?|wss?|ws?|ftp)://",
                },
                "delete": {
                    "type": "string",
                    "format": "uri",
                    "pattern": "^(https?|http?|wss?|ws?|ftp)://",
                },
            },
            "anyOf": [{"required": ["list"]}, {"required": ["instance"]},],
        },
        "buttons": {
            "type": "array",
            "items": {
                "type": "string",
                "enum": [
                    "save",
                    "saveandnew",
                    "saveandclose",
                    "refresh",
                    "new",
                    "delete",
                ],
            },
        },
        "custom_buttons": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "type": {
                        "type": "string",
                        "enum": ["hyperlink", "widget", "dropdown", "action"],
                    },
                    "key": {"type": "string"},
                    "label": {"type": "string"},
                    "icon": {"type": "string"},
                    "title": {"type": "string"},
                },
                "anyOf": [
                    {"required": ["type", "key", "label"]},
                    {"required": ["type", "key", "icon"]},
                ],
            },
        },
    },
    "required": ["identifier", "type"],
}

example_dict = {
    "identifier": "portfolio:strategy",
    "messages": [],
    "type": "list",
    "endpoints": {
        "list": "http://localhost:5000/api/portfolio/strategy/",
        "instance": "http://localhost:5000/api/portfolio/strategy/",
    },
    "buttons": ["save", "refresh"],
    "custom_buttons": [{"type": "widget", "key": "some_key", "label": "Some Label"}],
    "instance_display": [
        {"fields": ["title", ["isin", "ticker"], ["currency", "is_cash"]]}
    ],
    "list_display": [
        {"key": "title", "label": "Title"},
        {"key": "ticker", "label": "Ticker"},
        {"key": "currency", "label": "Currency"},
        {"key": "is_cash", "label": "Cash"},
    ],
    "list_formatting": None,
    "cell_formatting": None,
    "legends": None,
    "instance_buttons": ["refresh", "save", "delete"],
    "list_buttons": ["refresh", "new"],
    "custom_list_buttons": [],
    "custom_instance_buttons": [
        {"key": "strategyposition", "label": "Positions", "icon": "wb-icon-stats"},
        {
            "_buttons": [
                {
                    "key": "portfoliorole",
                    "label": "Portfolio Roles",
                    "icon": "wb-icon-person",
                }
            ]
        },
    ],
    "custom_list_instance_buttons": [
        {"key": "strategyposition", "label": "Positions", "icon": "wb-icon-stats"},
        {
            "_buttons": [
                {
                    "key": "portfoliorole",
                    "label": "Portfolio Roles",
                    "icon": "wb-icon-person",
                }
            ]
        },
    ],
    "list_widget_title": "Strategies",
    "instance_widget_title": "Strategy: {{title}}",
    "new_instance_widget_title": "New Strategy",
    "pagination_type": "cursor",
    "fields": {
        "id": {
            "key": "id",
            "label": "Id",
            "type": "primary_key",
            "required": False,
            "read_only": True,
        },
        "title": {
            "key": "title",
            "label": "Title",
            "type": "text",
            "required": True,
            "read_only": False,
        },
        "ticker": {
            "key": "ticker",
            "label": "Ticker",
            "type": "text",
            "required": True,
            "read_only": False,
        },
        "isin": {
            "key": "isin",
            "label": "ISIN",
            "type": "text",
            "required": True,
            "read_only": False,
        },
        "currency": {
            "key": "currency",
            "label": "Currency",
            "type": "select",
            "required": True,
            "read_only": False,
            "representation_key": "_currency",
            "endpoint": {
                "url": "http://localhost:5000/api/currency/currencyrepresentation/",
                "value_key": "id",
                "label_key": "{{key}} ({{symbol}})",
            },
        },
        "is_cash": {
            "key": "is_cash",
            "label": "Is cash",
            "type": "boolean",
            "required": True,
            "read_only": False,
        },
        "_additional_resources": {
            "key": "_additional_resources",
            "label": " additional resources",
            "type": "field",
            "required": False,
            "read_only": True,
        },
    },
    "filter_fields": {
        "title": {
            "label": "Title",
            "type": "text",
            "key": "title",
            "visible": False,
            "required": False,
        },
        "isin": {
            "label": "ISIN",
            "type": "text",
            "key": "isin",
            "visible": False,
            "required": False,
        },
        "ticker": {
            "label": "Ticker",
            "type": "text",
            "key": "ticker",
            "visible": False,
            "required": False,
        },
        "currency": {
            "label": "Currency",
            "type": "select",
            "key": "currency",
            "visible": False,
            "required": False,
            "multiple": True,
            "endpoint": {
                "url": "http://localhost:5000/api/currency/currency/",
                "value_key": "id",
                "label_key": "{{key}} ({{symbol}})",
            },
        },
        "is_cash": {
            "label": "Cash",
            "type": "boolean",
            "key": "is_cash",
            "visible": False,
            "required": False,
        },
        "parent_strategy": {
            "label": "Parent Strategy",
            "type": "select",
            "key": "parent_strategy",
            "visible": False,
            "required": False,
            "multiple": True,
            "endpoint": {
                "url": "http://localhost:5000/api/portfolio/strategyrepresentation/",
                "value_key": "id",
                "label_key": "{{title}} {{ticker}}",
            },
        },
    },
    "ordering_fields": ["title", "ticker"],
    "search_fields": ["title", "ticker"],
}
# example = json.dumps(example_dict)
# print(example)

validate(instance=example_dict, schema=schema)



from collections import defaultdict
from typing import List, NamedTuple, Optional, Union, Tuple, Any, Dict


class ColumnFormatting(NamedTuple):
    style: str
    condition: Optional[Tuple[str, Union[int, float, str]]] = None
    all_columns: bool = False

    def to_dict(self):
        rd = {"style": self.style, "all_columns": self.all_columns}
        if self.condition:
            rd["condition"] = list(self.condition)
        return rd


class RowIconCondition(NamedTuple):
    icon: str
    condition: Optional[Tuple[str, Union[int, float, str]]] = None

    def to_dict(self):
        rd = {"icon": self.icon}
        if self.condition:
            rd["condition"] = list(self.condition)
        return rd


class RowStyleCondition(NamedTuple):
    style: Dict
    condition: Optional[Tuple[str, Union[int, float, str]]] = None

    def to_dict(self):
        rd = {"style": self.style}
        if self.condition:
            rd["condition"] = list(self.condition)
        return rd


class RowFormatting(NamedTuple):
    column: str
    conditions: List[Union[RowIconCondition, RowStyleCondition]]

    def to_dict(self):
        rd = defaultdict(list, {"column": self.column})
        for condition in self.conditions:
            rd["conditions"].append(condition.to_dict())
        return rd


class Field(NamedTuple):
    key: str
    label: str
    formatting: List[ColumnFormatting] = []

    def to_dict(self):
        rd = defaultdict(list, {"key": self.key, "label": self.label})
        for formatting in self.formatting:
            rd["formatting"].append(formatting.to_dict())
        return rd


class LegendItem(NamedTuple):
    icon: str
    label: str
    value: Optional[Any] = None

    def to_dict(self):
        rv = {"icon": self.icon, "label": self.label}
        if self.value:
            tv["value"] = self.value
        return rv


class Legend(NamedTuple):
    items: List[LegendItem]
    label: Optional[str] = None
    key: Optional[str] = None

    def to_dict(self):
        tv = {"items": list()}

        if self.label:
            tv["label"] = self.label
        if self.key:
            tv["key"] = self.key

        for item in self.items:
            tv["items"].append(item.to_dict())

        return tv


class ListDisplay(NamedTuple):
    fields: List[Field]
    formatting: List[RowFormatting] = []
    legends: List[Legend] = []

    def to_dict(self):
        list_display = defaultdict(list)
        for field in self.fields:
            list_display["fields"].append(field.to_dict())

        for formatting in self.formatting:
            list_display["formatting"].append(formatting.to_dict())

        for legend in self.legends:
            list_display["legends"].append(legend.to_dict())

        return list_display


class FieldSet(NamedTuple):
    # This should include the recursive subtypes, but then mypy gets a recursion error
    fields: List

    def to_dict(self):
        rl = list()
        for field in self.fields:
            if isinstance(field, str):
                rl.append(field)
            else:
                rl.append(field.to_dict())
        return rl


class Section(NamedTuple):
    fields: FieldSet
    title: Optional[str] = None
    collapsed: bool = False

    def to_dict(self):
        rd = {"fields": self.fields.to_dict()}

        if self.title:
            rd["title"] = self.title

        if self.collapsed:
            rd["collapsed"] = self.collapsed

        return rd


class InstanceDisplay(NamedTuple):
    sections: List[Section]

    def to_dict(self):
        instance_display = list()
        for section in self.sections:
            instance_display.append(section.to_dict())
        return instance_display


# import json
# from pprint import pprint
# from bridger.enums import Operator

# i = InstanceDisplay(
#     sections=[
#         Section(fields=FieldSet(["a1", "a2", FieldSet(["a31", "a32"])])),
#         Section(
#             title="Section2",
#             collapsed=True,
#             fields=FieldSet(
#                 ["b1", "b2", FieldSet(["b31", FieldSet(["b311", "b312"])])]
#             ),
#         ),
#     ]
# )
# pprint(json.loads(json.dumps(i.to_dict())))

# l = ListDisplay(
#     fields=[
#         Field(key="title", label="Title"),
#         Field(
#             key="title",
#             label="Title",
#             formatting=[
#                 ColumnFormatting(style="bold", condition=(Operator.GREATER.value, 3))
#             ],
#         ),
#         Field(key="title", label="Title"),
#     ],
#     formatting=[
#         RowFormatting(
#             column="title",
#             conditions=[
#                 RowIconCondition(
#                     icon="wb-icon-something", condition=(Operator.EQUAL.value, "abc")
#                 ),
#                 RowStyleCondition(
#                     style="bold", condition=(Operator.UNEQUAL.value, "abc")
#                 ),
#             ],
#         )
#     ],
# )
# pprint(json.loads(json.dumps(l.to_dict())))



from django.db import models
from django.db.models.fields.related import ManyToManyRel, ManyToOneRel, OneToOneRel
from django_filters.rest_framework import FilterSet as DjangoFilterSet

from bridger.filters import fields


class FilterSet(DjangoFilterSet):
    FILTER_DEFAULTS = {
        models.BooleanField: {"filter_class": fields.BooleanFilter},
        models.NullBooleanField: {"filter_class": fields.BooleanFilter},
        models.CharField: {"filter_class": fields.CharFilter},
        models.TextField: {"filter_class": fields.CharFilter},
        models.SlugField: {"filter_class": fields.CharFilter},
        models.EmailField: {"filter_class": fields.CharFilter},
        models.FilePathField: {"filter_class": fields.CharFilter},
        models.URLField: {"filter_class": fields.CharFilter},
        models.GenericIPAddressField: {"filter_class": fields.CharFilter},
        models.CommaSeparatedIntegerField: {"filter_class": fields.CharFilter},
        models.DateField: {"filter_class": fields.DateFilter},
        # models.AutoField: {"filter_class": NumberFilter},
        # models.DateTimeField: {"filter_class": DateTimeFilter},
        # models.TimeField: {"filter_class": TimeFilter},
        # models.DurationField: {"filter_class": DurationFilter},
        # models.DecimalField: {"filter_class": NumberFilter},
        # models.SmallIntegerField: {"filter_class": NumberFilter},
        # models.IntegerField: {"filter_class": NumberFilter},
        # models.PositiveIntegerField: {"filter_class": NumberFilter},
        # models.PositiveSmallIntegerField: {"filter_class": NumberFilter},
        # models.FloatField: {"filter_class": NumberFilter},
        # models.UUIDField: {"filter_class": UUIDFilter},
        # Forward relationships
        models.OneToOneField: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "to_field_name": f.remote_field.field_name,
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        models.ForeignKey: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "to_field_name": f.remote_field.field_name,
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        models.ManyToManyField: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {"queryset": remote_queryset(f),},
        },
        # Reverse relationships
        OneToOneRel: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        ManyToOneRel: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {"queryset": remote_queryset(f),},
        },
        ManyToManyRel: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {"queryset": remote_queryset(f),},
        },
    }

    @classmethod
    def filter_for_lookup(cls, field, lookup_type):

        filter_class, params = super().filter_for_lookup(field, lookup_type)
        if hasattr(field, "verbose_name"):
            params["label"] = field.verbose_name

        return filter_class, params




from collections import defaultdict

from rest_framework.metadata import SimpleMetadata

from .enums import WidgetType
from .serializers import percent_decorator


class BridgerMetaData(SimpleMetadata):
    def determine_metadata(self, request, view):
        metadata = defaultdict(dict)

        metadata["type"] = view.get_widget_type(request=request)
        metadata["identifier"] = view.get_identifier(request=request)
        metadata["buttons"] = view.get_buttons(request=request)
        metadata["create_buttons"] = view.get_create_buttons(request=request)
        metadata["custom_buttons"] = view.get_custom_buttons(request=request)
        metadata["custom_instance_buttons"] = view.get_custom_instance_buttons(
            request=request
        )
        metadata["endpoints"] = view.get_endpoints(
            request=request, buttons=metadata["buttons"]
        )
        metadata["pagination"] = view.get_pagination(request=request)
        metadata["titles"] = view.get_titles(request=request)

        if metadata["type"] in [WidgetType.INSTANCE.value, WidgetType.LIST.value]:
            serializer_class = view.get_serializer_class()
            if "pk" in view.kwargs:
                metadata["pk"] = view.kwargs["pk"]

            metadata["list_display"] = view.get_list_display(request)
            metadata["instance_display"] = view.get_instance_display(request)

            metadata["fields"] = view.get_fields(request)
            for key, value in serializer_class.get_decorators():
                metadata["fields"][key]["decorators"].append(value)
            for key in serializer_class.get_percent_fields():
                metadata["fields"][key]["type"] = "percent"
                metadata["fields"][key]["decorators"].append(percent_decorator)

        metadata["search_fields"] = view.get_search_fields(request)
        metadata["ordering_fields"] = view.get_ordering_fields(request)
        metadata["filter_fields"] = {k: v for k, v in view.get_filter_fields(request)}

        # TODO: Custom Buttons

        return metadata



from rest_framework import serializers

from .mixins import BridgerSerializerFieldMixin
from .types import BridgerType


class PrimaryKeyField(BridgerSerializerFieldMixin, serializers.IntegerField):
    field_type = BridgerType.PRIMARY_KEY.value

    def __init__(self, *args, **kwargs):
        kwargs["read_only"] = True
        kwargs["required"] = False
        super().__init__(*args, **kwargs)


class PrimaryKeyCharField(BridgerSerializerFieldMixin, serializers.CharField):
    field_type = BridgerType.PRIMARY_KEY.value

    def __init__(self, *args, **kwargs):
        kwargs["read_only"] = True
        kwargs["required"] = False
        super().__init__(*args, **kwargs)



from rest_framework.reverse import reverse

from bridger.serializers.fields.related import ListSerializer


class RepresentationSerializerMixin:
    """
    Adds basic functionality for Representation Serializer
    """

    class Meta:
        list_serializer_class = ListSerializer

    def __init__(self, *args, **kwargs):
        self.label_key = kwargs.pop("label_key", self.label_key)
        super().__init__(*args, **kwargs)

    def __new__(cls, *args, **kwargs):
        if "read_only" not in kwargs:
            kwargs["read_only"] = True
        return super().__new__(cls, *args, **kwargs)

    def get_filter_params(self, request):
        if hasattr(self, "filter_params"):
            return self.filter_params
        return None

    def get_representation(self, request, field_name):

        url = reverse(self.endpoint, request=request)

        filter_params = self.get_filter_params(request)
        if filter_params:
            url = f"{url}?{urlencode(filter_params)}"

        representation = {
            "representation_key": field_name,
            "related_key": self.source,
            "endpoint": {
                "url": url,
                "value_key": self.value_key,
                "label_key": self.label_key,
            },
        }
        if self.help_text:
            representation["help_text"] = self.help_text

        return representation



from typing import Union

from enum import Enum


class Button(Enum):
    # List and Instance
    REFRESH = "refresh"
    RESET = "reset"
    NEW = "new"
    DELETE = "delete"

    # Instance
    SAVE = "save"
    SAVE_AND_CLOSE = "saveandclose"
    SAVE_AND_NEW = "saveandnew"

    # Custom
    DROPDOWN = "dropdown"
    HYPERLINK = "hyperlink"
    WIDGET = "widget"
    ACTION = "action"


class WidgetType(Enum):
    LIST = "list"
    INSTANCE = "instance"
    CHART = "chart"


class Operator(Enum):
    EQUAL = "=="
    UNEQUAL = "!="
    GREATER = ">"
    GREATER_EQUAL = ">="
    LESS = "<"
    LESS_EQUAL = "<="


class Unit(Enum):
    FRACTION = "fr"
    REM = "rem"
    PIXEL = "px"

    def unit(self, _value: Union[float, str, int]):
        assert isinstance(
            _value, (float, str, int)
        ), f"_value needs to be one of str, float or int"

        return f"{float(_value)}{self.value}"


class AuthType(Enum):
    NONE = "NONE"
    JWT = "JWT"



from .fields import AdditionalResourcesField, HyperlinkField, register_resource

from .boolean import BooleanField
from .choice import ChoiceField
from .datetime import DateTimeField, DateField, TimeField
from .file import ImageField, FileField
from .fsm import FSMStatusField
from .json import JSONField, JSONTextEditorField
from .list import ListField
from .number import IntegerField, DecimalField, FloatField
from .primary_key import PrimaryKeyField, PrimaryKeyCharField
from .related import PrimaryKeyRelatedField, ListSerializer
from .text import CharField, StringRelatedField, TextField

from .other import StarRatingField, RangeSelectField



from django.conf import settings
from rest_framework.views import APIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.reverse import reverse

from .enums import AuthType
from .menus import default_registry


class Menu(APIView):
    # permission_classes = [IsAuthenticated]

    def get(self, request: Request) -> Response:
        return Response(default_registry.to_dict(request))


class Profile(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request: Request) -> Response:
        return Response(
            {
                "config": reverse(
                    "bridger:frontenduserconfiguration-list", request=request
                ),
                "image": "TODO: Here comes the URL",
            }
        )


class Config(APIView):
    def get(self, request: Request) -> Response:
        auth = getattr(settings, "BRIDGER_AUTH")

        return Response(
            {
                # "authentication": auth(request),
                "authentication": {"type": AuthType.NONE.name},
                "notification": {},
                "profile": reverse("bridger:profile", request=request),
                # "profile": {},
                "menu": reverse("bridger:menu", request=request),
            }
        )



from .fields import (
    BooleanFilter,
    CharFilter,
    ChoiceFilter,
    DateFilter,
    DefaultDateRangeFilterValues,
    ModelChoiceFilter,
    ModelMultipleChoiceFilter,
    MultipleChoiceFilter,
)
from .filterset import FilterSet




import logging

import django_filters
from django.db.models import QuerySet
from django.urls.exceptions import NoReverseMatch
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import filters, status, viewsets
from rest_framework.decorators import api_view, permission_classes
from rest_framework.mixins import ListModelMixin
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response

from .fsm.mixins import FSMViewSetMixin
from .filters import BooleanFilter, ModelChoiceFilter
from .mixins import MetadataMixin
from .pagination import CursorPagination

logger = logging.getLogger(__name__)


class RepresentationModelViewSet(MetadataMixin, viewsets.ReadOnlyModelViewSet):
    """A Representation View that is used for serializing related fields"""

    filter_backends = [filters.OrderingFilter]
    pagination_class = CursorPagination

    ordering_fields = ordering = ["id"]


class ModelViewSet(MetadataMixin, FSMViewSetMixin, viewsets.ModelViewSet):
    """A Model View that is used to serializer models"""

    filter_backends = [filters.OrderingFilter]
    pagination_class = CursorPagination

    ordering_fields = ordering = ["id"]

    @classmethod
    def get_model(cls):
        try:
            if hasattr(cls, "queryset"):
                return cls.queryset.model
            elif hasattr(cls, "serializer_class"):
                return cls.serializer_class.Meta.model
            else:
                return None
        except AttributeError:
            return None

    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        serialized_content = {"instance": serializer.data}

        if hasattr(self, "get_messages"):
            serialized_content["messages"] = self.get_messages(request)

        return Response(serialized_content)


class ChartViewSet(MetadataMixin, ListModelMixin, viewsets.ViewSet):
    """A List View that is used for creating plotly charts"""

    CHART_DISPLAY = True
    filter_backends = [DjangoFilterBackend]

    def filter_queryset(self, queryset: QuerySet) -> QuerySet:
        for backend in list(self.filter_backends):
            queryset = backend().filter_queryset(self.request, queryset, self)
        return queryset


@api_view(["GET"])
@permission_classes([IsAuthenticated])
def api_endpoints_root(request: Request, format: str = None) -> Response:
    """Returns a list of API-ROOT endpoints specified in BRIDGER_ENDPOINTS"""

    endpoints = dict()
    endpoints["config"] = reverse("bridger:config", request=request, format=format)

    if hasattr(settings, "BRIDGER_ENDPOINTS"):
        for bridger_endpoint in settings.BRIDGER_ENDPOINTS:
            try:
                endpoints[bridger_endpoint] = reverse(
                    f"{bridger_endpoint}:api-root", request=request, format=format
                )
            except NoReverseMatch:
                endpoints[bridger_endpoint] = reverse(
                    f"API-ROOT of {bridger_endpoint} is not yet implemented."
                )

    return Response(endpoints, status=status.HTTP_200_OK)
